@@ -7,7 +7,6 @@ from gettext import gettext as _
 from logging import getLogger
 
 from astersay.agi import Agi
-from astersay.exceptions import AgiHangupError
 from astersay.conf import settings
 from astersay.dialog import Dialog
 
@@ -64,9 +63,6 @@ class App:
             self.stream = stream
             self.start()
             self.stop()
-        except AgiHangupError as e:
-            logger.info(str(e))
-            self.stop()
         except Exception as e:
             self.stop()
             logger.error(str(e), exc_info=e)
