@@ -91,7 +91,7 @@ class VoskRecognizer(BaseBackend):
             dialog.stop()
             return
 
-        listener = Listener(dialog.stream)
+        listener = Listener(dialog.stream, dialog.amplitude)
         self.listener = listener
         # Максимальный размер потокового файла для Vosk задаётся настройками.
         listener.max_size = self.max_size
