@@ -117,7 +117,7 @@ class YandexRecognizer(YandexBaseBackend):
         logger_info = self.logger.info
         logger_info(_('Начинается распознавание речи для диалога.'))
 
-        listener = Listener(dialog.stream)
+        listener = Listener(dialog.stream, dialog.amplitude)
         self.listener = listener
         # Максимальный размер потокового файла для Яндекса - 10 мегабайт.
         listener.max_size = 10485760
