@@ -34,6 +34,9 @@ class TextBuffer(list):
         self.key = key
         # Список индексов зафиксированных текстов.
         self.fixed = []
+        # Время финализированного текста, должно сбрасываться предварительным
+        # распознаванием.
+        self.final = None
 
     def get_text(self, mode=1):
         """
@@ -51,6 +54,12 @@ class TextBuffer(list):
             L = ['%s: %s' % (i + 1, x) for i, x in enumerate(self) if x]
         return '\n'.join(L)
 
+    @property
+    def seconds_after_final(self):
+        if self.final:
+            return time() - self.final
+        return 0
+
 
 def _make_say_params(saydata, variables):
     """Возвращает готовые аргументы для метода say()."""
@@ -60,6 +69,8 @@ def _make_say_params(saydata, variables):
         params['voice'] = saydata['voice'] % variables
     elif 'file' in saydata:
         params['voice'] = splitext(saydata['file'])[0] % variables
+    elif 'texts' in saydata:
+        params['text'] = '|'.join(saydata['texts']) % variables
     else:
         params['text'] = saydata['text'] % variables
 
@@ -139,11 +150,20 @@ class Dialog:
     _voice = None
     _voice_duration = 0
     _voice_nonblocking = False
+    # RMS-значение данных тихого участка звукового потока используется для
+    # рассчёта тишины канала и может уменьшаться в процессе работы.
+    _silence_rms = 1000
+    # Момент начала тишины в Unix-time (целые секунды).
+    _silence_from = None
+    # Флаг, при установке которого сбрасывается счётчик секунд молчания.
+    _reset_silence = False
 
     def __init__(self, agi, settings, stream):
         self.agi = agi
         self.settings = settings
         self.stream = stream
+        self.amplitude = []
+        self.silence_seconds = 0
 
         self.synthesizer = settings.get_synthesizer()
         self.recognizer = settings.get_recognizer()
@@ -174,6 +194,7 @@ class Dialog:
         text = text.strip()
         if not text:
             return
+        text_buffer.final = None
         # Первые передварительные результаты не принимаются до заданного
         # ограничения.
         if not fixed:
@@ -189,6 +210,7 @@ class Dialog:
         if fixed:
             text_buffer.fixed.append(len(text_buffer) - 1)
             text_buffer.append('')
+            text_buffer.final = time()
 
     def text_from_text_buffer(self, mode=1, text_buffer=None):
         """
@@ -292,14 +314,20 @@ class Dialog:
             log_debug(_('Голос отправляется в блокирующем режиме.'))
 
         def clear_voice():
-            rest = duration
-            while rest > 0 and self._voice == voice:
-                sleep(0.5)
-                rest -= 0.5
             if self._voice == voice:
                 self._voice = None
                 self._voice_duration = 0
                 log_debug(_('Голос %s завершён.'), voice)
+                # После завершения любой речи робота мы должны сбросить счётчик
+                # молчания.
+                self._reset_silence = True
+
+        def runner():
+            rest = duration
+            while rest > 0 and self._voice == voice:
+                sleep(0.5)
+                rest -= 0.5
+            clear_voice()
 
         try:
             self.agi.stream_file(voice, escape_digits=escape_digits)
@@ -310,14 +338,58 @@ class Dialog:
             log_info(_('Голос не отправлен. Связь разорвана.'))
             return
 
-        thread = Thread(target=clear_voice, daemon=True)
+        thread = Thread(target=runner, daemon=True)
         thread.start()
 
         self._voice_nonblocking = nonblocking
         if not nonblocking:
             log_debug(_('Останавливаем основной поток на %s секунд.'), duration)
             # Останавливаем основной поток.
-            sleep(duration)
+            sleep(duration + 0.1)
+            # Запускаем гарантированную очистку.
+            clear_voice()
+
+    def say_silence(self, conf, index, variables):
+        try:
+            s = conf['phrases'][index]
+            index += 1
+        except IndexError:
+            s = conf['phrases'][0]
+            index = 1
+        saydata = {'text': s}
+        self.say(**_make_say_params(saydata, variables))
+        return index
+
+    def calc_silence_seconds(self):
+        silence_rms = self._silence_rms
+        # Граница тихого участка - это 200% от минимума.
+        # Вроде бы смысла нет иметь границу ниже 500 единиц RMS.
+        if silence_rms > 250:
+            # Уменьшаем минимальное значение.
+            for second, rms in self.amplitude:
+                if rms < 250:
+                    self._silence_rms = silence_rms = 250
+                    log_debug(_('Установлен минимальный RMS=250(%d).'), rms)
+                    break
+                if rms < silence_rms:
+                    self._silence_rms = silence_rms = rms
+                    log_debug(_('Установлен минимальный RMS=%d.'), rms)
+
+        border = silence_rms * 2
+        silence_from = self._silence_from
+        for second, rms in reversed(self.amplitude):
+            if rms > border:
+                self._silence_from = silence_from = second + 1
+                break
+        if silence_from is None:
+            value = 0
+        else:
+            seconds = int(time()) - silence_from
+            value = seconds if seconds > 0 else 0
+        self.silence_seconds = value
+        if value:
+            log_debug(_('Тишина %r секунд.'), value)
+        return value
 
     def has_start_recognize(self):
         if self.is_closed:
@@ -325,12 +397,40 @@ class Dialog:
         agi = self.agi
         return not agi.is_sighup and not agi.is_hungup
 
+    @property
+    def is_stopped(self):
+        if self.is_closed:
+            return True
+        agi = self.agi
+        return agi.is_sighup or agi.is_hungup
+
     def start(self):
         if getattr(self, '_recognizer_thread', None):
             raise RuntimeError(_('Другой поток распознавания уже запущен.'))
         if not self.text_buffers:
             raise RuntimeError(_('Начальный текстовый буфер удалён.'))
 
+        def calc_silence():
+            # От начала и до конца беседы поток голоса абонента анализируется
+            # на тишину каждую секунду.
+            while self.has_start_recognize():
+                if self._reset_silence:
+                    # self.amplitude.clear()
+                    self._silence_from = int(time())
+                    self._reset_silence = False
+                    log_debug(_('Сброс счётчика молчания.'))
+                sleep(1)
+                self.calc_silence_seconds()
+            thread = self._calc_silence_thread
+            self._calc_silence_thread = None
+            log_debug(_('Завершён поток подсчёта тишины %s.'), thread)
+
+        log_debug(_('Запуск отдельного потока подсчёта тишины.'))
+        thread = Thread(target=calc_silence, daemon=True)
+        thread.start()
+        log_debug(_('Запущен поток подсчёта тишины %s.'), thread)
+        self._calc_silence_thread = thread
+
         def listen():
             # От начала и до конца беседы поток голоса абонента анализируется.
             # Причём, когда объём потока превышает разрешённые 10 мегабайт
@@ -380,7 +480,7 @@ class Dialog:
             log_info(_('Абонент повесил трубку до окончания диалога.'))
             hangup = True
         except AgiError as e:
-            error = _('Ошибка в AGI.')
+            error = _('Ошибка в AGI при исполнении диалога.')
             log_error(error, exc_info=e)
         except Exception as e:
             error = _('Ошибка в сценарии диалога.')
@@ -388,13 +488,22 @@ class Dialog:
         self.stop()
 
         agi = self.agi
-        for key, value in self.export.items():
-            if key.startswith('_'):
-                continue
-            value = str(value)
-            log_debug(
-                _('Экспорт %(key)s=%(value)s'), {'key': key, 'value': value})
-            agi.set(key, value)
+        try:
+            for key, value in self.export.items():
+                if key.startswith('_'):
+                    continue
+                value = str(value)
+                log_debug(_('Экспорт %(key)s=%(value)s'), {
+                    'key': key, 'value': value})
+                agi.set(key, value)
+        except AgiHangupError:
+            log_info(
+                _('Абонент повесил трубку до окончания передачи переменных.'))
+            hangup = True
+        except AgiError as e:
+            error = _('Ошибка в AGI при передаче переменных.')
+            log_error(error, exc_info=e)
+
         filename = self.settings.export_dialog(
             self.agi.params, self.export, self.text_buffers,
             hangup=hangup,
@@ -428,7 +537,8 @@ class Dialog:
             'chain': self.chain,
             'text_buffers': self.text_buffers,
         }
-        if isinstance(result, (str, list, tuple, int, float)):
+        if result is None or \
+                isinstance(result, (str, bool, list, tuple, int, float)):
             plugin_data['result'] = result
         else:
             plugin_data['result'] = str(result)
@@ -451,6 +561,30 @@ class Dialog:
             log_debug('< PLUGIN %s', res.stdout)
             plugin_data = json.load(open(tempfile_name))
             variables.update(plugin_data['variables'])
+            # Предварительный фоновый синтез текстов.
+            synthesize = plugin_data.get('synthesize', [])
+            if synthesize and isinstance(synthesize, list):
+                def make_voices():
+                    log_debug(
+                        _('Предварительный синтез для %d текста(ов).'),
+                        len(synthesize)
+                    )
+                    for txt in synthesize:
+                        self.synthesizer.make_voice(str(txt), early=True)
+                    log_debug(_('Предварительный синтез речи завершён.'))
+                t = Thread(target=make_voices, daemon=True)
+                t.start()
+            # Логгирование каких-либо строк из плагина.
+            loggers = (
+                ('debug', log_debug),
+                ('info', log_info),
+                ('warnings', log_warning),
+                ('errors', log_error),
+            )
+            for _name, _logger in loggers:
+                strings = plugin_data.get(_name, [])
+                for string in strings:
+                    _logger('< PLUGIN %s', string)
         tempfile.close()
         return res, plugin_data
 
@@ -549,9 +683,9 @@ class Dialog:
                 except IndexError:
                     process = None
             log_debug('    NEXT %s', process)
-            if say_debug and process:
-                saydata = {'text': _('Следующий скрипт %(name)r') % process}
-                self.say(**_make_say_params(saydata, variables))
+            # if say_debug and process:
+            #     saydata = {'text': _('Следующий скрипт %(name)r') % process}
+            #     self.say(**_make_say_params(saydata, variables))
 
             # Восстанавливаем константы. Это проще, чем городить лес кода.
             variables.update(constants)
@@ -600,6 +734,10 @@ class Dialog:
         if not self._voice:
             log_debug(_('Нечего прерывать, голос уже закончился.'))
             return
+        # Сначала должен среагировать поток озвучки голоса, а затем отправляем
+        # команду остановки.
+        self._voice = None
+        self._voice_duration = 0
         try:
             self.agi.stream_file(stop_voice)
         except AgiSigpipeError:
@@ -609,8 +747,6 @@ class Dialog:
         except Exception as e:
             self.is_closed = True
             raise e
-        self._voice = None
-        self._voice_duration = 0
         log_debug(_('Прерывание воспроизведения выполнено.'))
 
     def _script_capture(self, script, variables):
@@ -630,6 +766,7 @@ class Dialog:
         assert mode in (1, 2, 3)
         # Для капчуринга всегда создаём новый буфер.
         self.make_text_buffer()
+        text_buffer = self.get_last_text_buffer()
 
         saydata = script.get('say_start')
         if saydata:
@@ -638,6 +775,7 @@ class Dialog:
             self.say(**_make_say_params(saydata, variables))
             _make_pause(pauses, 'after_say_start')
             self._agi_execute(commands, 'after_say_start')
+        say_start_duration = self._voice_duration
 
         maxtime = float(script.get('maxtime', 600))
         minlength = int(script.get('minlength', 10))
@@ -651,45 +789,41 @@ class Dialog:
 
         func = self.capture_text
         speech, stop = func(stopwords, mode=mode)
-        # Переменная для определения молчания.
-        _speech = speech
         if stop:
             log_debug(_('Первая попытка успешна. Текст: %s'), speech)
         else:
             log_debug(_('Первая попытка провалилась. Текст: %s'), speech)
         # В блокирующем режиме голоса его длина уже будет сброшена в 0.
-        maxtime += self._voice_duration
+        maxtime += say_start_duration
 
         second = 0
         # Количество секунд тишины.
-        silence_second = 0
         while not stop and second < maxtime:
             log_debug(_('Стоп-слова пока нет, ждём %d-ю секунду.'), second)
             second += 0.5
             sleep(0.5)
-            silence_second += 0.5
             speech, stop = func(stopwords, mode=mode)
             if stop:
                 log_debug('Сработало ограничение по стоп-слову.')
             else:
-                _expire = silence_second >= silence
+                silence_seconds = self.silence_seconds
+                _expire = silence_seconds >= silence
                 _size = len(speech) >= minlength
-                stop = _expire and _size
+                _final = text_buffer.seconds_after_final > 1
+                stop = _expire and _size and _final
                 if _expire:
                     log_debug('Сработало ограничение по тишине.')
                     if _size:
                         log_debug('Сработало ограничение по длине.')
+                        if _final:
+                            log_debug('Сработало ограничение по финалу '
+                                      'распознавания.')
+                        else:
+                            log_debug('Ожидаем финальное распознавание.')
                     else:
                         log_debug('Длина текста недостаточна для завершения.')
-            if _speech != speech:
-                # log_debug('old speech %r', _speech)
-                # log_debug('new speech %r', speech)
-                silence_second = 0
-                _speech = speech
-            else:
-                log_debug(_('Тишина %r секунд.'), silence_second)
-                pass
-            if self.is_closed:
+
+            if self.is_stopped:
                 log_debug(_('Диалог завершился до приёма данных.'))
                 break
 
@@ -755,6 +889,14 @@ class Dialog:
             self.say(**_make_say_params(saydata, variables))
             _make_pause(pauses, 'after_say_start')
             self._agi_execute(commands, 'after_say_start')
+        say_start_duration = self._voice_duration
+
+        say_silence = script.get('say_silence')
+        if say_silence:
+            say_silence_time = say_silence['time']
+        else:
+            say_silence_time = 0
+        say_silence_index = 0
 
         attempts = script.get('attempts') or {}
         attempts_count = abs(int(attempts.get('count', 0))) or 1
@@ -771,16 +913,23 @@ class Dialog:
             _iter_time = attempts_iter_time
             if attempt == 0:
                 # В блокирующем режиме голоса его длина уже будет сброшена в 0.
-                _iter_time += self._voice_duration
+                _iter_time += say_start_duration
             second = 0
             while not names and second < _iter_time:
                 log_debug(_('Ответа пока нет, ждём %d-ю секунду.'), second)
                 second += attempts_iter_pause
                 sleep(attempts_iter_pause)
                 names, speech = func(mode=mode)
-                if self.is_closed:
+                if self.is_stopped:
                     log_debug(_('Диалог завершился до приёма данных.'))
                     break
+                # Обработка молчания ягнят.
+                if say_silence_time and (attempt > 0 or second > say_start_duration):
+                    ss = self.silence_seconds
+                    if second and ss and ss % say_silence_time == 0:
+                        log_debug(_('Кратность %r.'), say_silence_time)
+                        say_silence_index = self.say_silence(
+                            say_silence, say_silence_index, variables)
             if names:
                 length_names = int(script.get('length_names', 2))
                 answer = ' '.join(names[:length_names])
@@ -856,6 +1005,14 @@ class Dialog:
             self.say(**_make_say_params(saydata, variables))
             _make_pause(pauses, 'after_say_start')
             self._agi_execute(commands, 'after_say_start')
+        say_start_duration = self._voice_duration
+
+        say_silence = script.get('say_silence')
+        if say_silence:
+            say_silence_time = say_silence['time']
+        else:
+            say_silence_time = 0
+        say_silence_index = 0
 
         attempts = script.get('attempts') or {}
         attempts_count = abs(int(attempts.get('count', 0))) or 1
@@ -873,16 +1030,23 @@ class Dialog:
             _iter_time = attempts_iter_time
             if attempt == 0:
                 # В блокирующем режиме голоса его длина уже будет сброшена в 0.
-                _iter_time += self._voice_duration
+                _iter_time += say_start_duration
             second = 0
             while not match and second < _iter_time:
                 log_debug(_('Ответа пока нет, ждём %d-ю секунду.'), second)
                 second += attempts_iter_pause
                 sleep(attempts_iter_pause)
                 match, speech = func(pattern, mode=mode)
-                if self.is_closed:
+                if self.is_stopped:
                     log_debug(_('Диалог завершился до приёма данных.'))
                     break
+                # Обработка молчания ягнят.
+                if say_silence_time and (attempt > 0 or second > say_start_duration):
+                    ss = self.silence_seconds
+                    if second and ss and ss % say_silence_time == 0:
+                        log_debug(_('Кратность %r.'), say_silence_time)
+                        say_silence_index = self.say_silence(
+                            say_silence, say_silence_index, variables)
             if match:
                 answer = match.group()
                 log_debug(_('Ответ: %r'), answer)
@@ -957,6 +1121,14 @@ class Dialog:
             self.say(**_make_say_params(saydata, variables))
             _make_pause(pauses, 'after_say_start')
             self._agi_execute(commands, 'after_say_start')
+        say_start_duration = self._voice_duration
+
+        say_silence = script.get('say_silence')
+        if say_silence:
+            say_silence_time = say_silence['time']
+        else:
+            say_silence_time = 0
+        say_silence_index = 0
 
         attempts = script.get('attempts') or {}
         attempts_count = abs(int(attempts.get('count', 0))) or 1
@@ -973,16 +1145,23 @@ class Dialog:
             _iter_time = attempts_iter_time
             if attempt == 0:
                 # В блокирующем режиме голоса его длина уже будет сброшена в 0.
-                _iter_time += self._voice_duration
+                _iter_time += say_start_duration
             second = 0
             while not response and second < _iter_time:
                 log_debug(_('Ответа пока нет, ждём %d-ю секунду.'), second)
                 second += attempts_iter_pause
                 sleep(attempts_iter_pause)
                 response, speech = func(filename, mode=mode)
-                if self.is_closed:
+                if self.is_stopped:
                     log_debug(_('Диалог завершился до приёма данных.'))
                     break
+                # Обработка молчания ягнят.
+                if say_silence_time and (attempt > 0 or second > say_start_duration):
+                    ss = self.silence_seconds
+                    if second and ss and ss % say_silence_time == 0:
+                        log_debug(_('Кратность %r.'), say_silence_time)
+                        say_silence_index = self.say_silence(
+                            say_silence, say_silence_index, variables)
             if response:
                 answer = response
                 log_debug(_('Ответ: %r'), answer)
@@ -1041,6 +1220,13 @@ class Dialog:
         if script.get('new_buffer', False):
             self.make_text_buffer()
 
+        plugin_name = script.get('plugin_before')
+        if plugin_name:
+            res, plugin_data = self.execute_plugin(plugin_name, '', '', variables)
+            result = plugin_data.get('result')
+        else:
+            result = True
+
         saydata = script.get('say_start')
         if saydata:
             _make_pause(pauses, 'before_say_start')
@@ -1049,8 +1235,13 @@ class Dialog:
             _make_pause(pauses, 'after_say_start')
             self._agi_execute(commands, 'after_say_start')
 
+        plugin_name = script.get('plugin_after')
+        if plugin_name:
+            res, plugin_data = self.execute_plugin(plugin_name, result, '', variables)
+            result = plugin_data.get('result')
+
         _make_pause(pauses, 'after')
         self._agi_execute(commands, 'after')
 
         speech = self.text_from_text_buffer()
-        return True, speech
+        return result, speech
